% Demo of CMDL-Dis (Cross-view Multi-Level Dictionary Learning with Discriminative Regularization) on VIPeR dataset
%
% Author: Sheng Li (shengli@ece.neu.edu)
%
% [1] Sheng Li, Ming Shao, Yun Fu: Person Re-Identification by Cross-View 2 Multi-Level Dictionary Learning. IEEE Trans. PAMI, 2018.
% [2] Sheng Li, Ming Shao, Yun Fu: Cross-View Projective Dictionary Learning for Person Re-Identification. IJCAI 2015: 2155-2161.

close all;

%%--Current working directory
project_dir = strcat(pwd, '/');

addpath('Code/');
%%--We need to call some functions from the "salience_reid" package, which
%%--can be downloaded at: https://github.com/Robert0812/salience_reid
% Rui Zhao, Wanli Ouyang, and Xiaogang Wang. Unsupervised Salience Learning
% for Person Re-identification. In IEEE Conference of Computer Vision and
% Pattern Recognition (CVPR), 2013.
addpath(genpath([project_dir 'Code/salience_reid-master/']));


%%--We also need the package OMP-Box v10 (http://www.cs.technion.ac.il/~ronrubin/Software/ompbox10.zip)
addpath(genpath([project_dir 'Code/ompbox10/']));

%%--Random Trail ID (from 1 to 10)--%%
testTrail = 1;

%%--Setting Parameters--%%
para = [];
para.patchsize = 10;
para.gridstep = 4;
para.numTr = 200;           %% Number of training pairs
para.useMask = 1;           %% Use mask or not
para.msk_thr = 0.2;
para.sigma1 = 2.8;
para.nor = 2;
para.alpha = 0.7;

phiFun = @(x, m1, par) (exp(-double(x).^2/par.sigma1^2).*m1).^par.nor;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%--Stage 0: Load the VIPeR dataset--%%

data_dir = [project_dir 'Data/'];
[trnS1, trnS2, S1, S2, nx, ny, dim] =  load_VIPeR(data_dir, testTrail, para);                 %% Load the VIPeR dataset
    
%%--Set gallary and probe sets--%%
trnSg = trnS1;
trnSp = trnS2;
Sg = S1;
Sp = S2;
    
%%--Load training data (features)--%%
% Note: our package doesn't include the features due to the large file size,
% please use the "salience_reid" package to create the features
feat_dir = [project_dir 'Data/dfeat/'];
Nr = para.numTr;
nynx = ny*nx;
patchTrainGal = zeros(dim, nynx*Nr);
patchTrainPrb = zeros(dim, nynx*Nr);
for i = 1:Nr
    load([feat_dir, 'feat', num2str(trnSg(i)), '.mat']);
    patchTrainGal(:, (i-1)*nynx+1:i*nynx) = densefeat;
    load([feat_dir, 'feat', num2str(trnSp(i)), '.mat']);
    patchTrainPrb(:, (i-1)*nynx+1:i*nynx) = densefeat;
end
    
patchTrainGal = normcol_equal(patchTrainGal);
patchTrainPrb = normcol_equal(patchTrainPrb);

%%--Load test data (features)--%%
gsize = 316;
testGal = zeros(dim, nynx*gsize);
testPrb = zeros(dim, nynx*gsize);
for i = 1:gsize
    load([feat_dir, 'feat', num2str(Sg(i)), '.mat']);
    testGal(:, (i-1)*nynx+1:i*nynx) = densefeat;
    load([feat_dir, 'feat', num2str(Sp(i)), '.mat']);
    testPrb(:, (i-1)*nynx+1:i*nynx) = densefeat;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%--Stage 1: Patch-level Dictionary Learning--%%
opt = [];
opt.dsize = 50;
opt.lambda = 0.001;
opt.gamma = 1e-4;
opt.iter = 30;
[DL1, DL2] = myDPL_PatchLevel(patchTrainGal, patchTrainPrb, opt);
DL1 = normcol_equal(DL1);
DL2 = normcol_equal(DL2);

%%--Coding test samples--%%
Xte_gal = projTestData(DL1,testGal,opt.dsize);
Xte_prb = projTestData(DL2,testPrb,opt.dsize);

%%%----Evaluation-----%%%
for i = 1:gsize
    if mod(i,10)==0
        fprintf('Now processing sample %d\n',i);
    end
    for j = 1:gsize
        xprbi = Xte_prb(:,(i-1)*nynx+1:i*nynx);
        xgalj = Xte_gal(:,(j-1)*nynx+1:j*nynx);
        [pwmap_prb(i, j).forward, pwmap_prb(i, j).fpos, ~] = ...
            mutualmap(xprbi, xgalj);
    end
end
pwmap_tst_cell = struct2cell(pwmap_prb);
D_cell = squeeze(pwmap_tst_cell(1, :, :));
if para.useMask
    load('Data/posemask_viper.mat');
    mask_prb = squeeze(mat2cell(mask(:, :, Sp) >= para.msk_thr, ny, nx, ones(1, length(Sp))));
    mp_cell = repmat(mask_prb, 1, gsize);
end
phi = cell(gsize, gsize);
for i = 1:numel(D_cell)
    phi{i} = phiFun(D_cell{i}, mp_cell{i}, para);
end
patchLevelDist = cellfun(@(x) dot(ones(1, numel(x)), x(:)), phi);
patchLevelDist = patchLevelDist';
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%--Stage 2: Image-level Dictionary Learning--%%
imageTrainGal = zeros(size(patchTrainGal,1)*300,size(patchTrainGal,2)/300);
for i = 1:size(imageTrainGal,2)
    tmp = patchTrainGal(:,(i-1)*300+1:i*300);
    imageTrainGal(:,i) = tmp(:);
end
imageTrainPrb = zeros(size(patchTrainPrb,1)*300,size(patchTrainPrb,2)/300);
for i = 1:size(imageTrainPrb,2)
    tmp = patchTrainPrb(:,(i-1)*300+1:i*300);
    imageTrainPrb(:,i) = tmp(:);
end

options=[];
options.ReducedDim = 190;
[Pca1,~] = PCA(imageTrainGal',options);
Y1 = imageTrainGal'*Pca1;

options=[];
options.ReducedDim = 190;
[Pca2,~] = PCA(imageTrainPrb',options);
Y2 = imageTrainPrb'*Pca2;

clear imageTrainGal imageTrainPrb

Y1 = normcol_equal(Y1');
Y2 = normcol_equal(Y2');

opt = [];
opt.dsize = 50;
opt.lambda1 = 0.7;
opt.lambda2 = 0.001;
opt.lambda3 = 0.001;
opt.lambda4 = 0.1;
opt.gamma = 1e-4;
opt.iter = 30;

R = ones(size(Y1,2)) ./ size(Y1,2);
for i = 1:size(R,2)
    R(i,i) = 0;
end
opt.R = R;

[DH1, DH2] = myDPL_ImageLevel_Dis(Y1, Y2, opt);

DH1 = normcol_equal(DH1);
DH2 = normcol_equal(DH2);

%%--Evaluation--%%%
imageTestGal = zeros(size(testGal,1)*300,size(testGal,2)/300);
for i = 1:size(imageTestGal,2)
    tmp = testGal(:,(i-1)*300+1:i*300);
    imageTestGal(:,i) = tmp(:);
end
imageTestPrb = zeros(size(testPrb,1)*300,size(testPrb,2)/300);
for i = 1:size(imageTestPrb,2)
    tmp = testPrb(:,(i-1)*300+1:i*300);
    imageTestPrb(:,i) = tmp(:);
end
Yte_gal = projTestData(DH1,Pca1'*imageTestGal,opt.dsize);
Yte_prb = projTestData(DH2,Pca2'*imageTestPrb,opt.dsize);
clear imageTestGal imageTestPrb
    
imageLevelDist = zeros(gsize);
for i = 1:gsize
    for j = 1:gsize
        imageLevelDist(i,j) = (Yte_gal(:,i)'*Yte_prb(:,j))/(norm(Yte_gal(:,i))*norm(Yte_prb(:,j)));
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%--Stage 3: Part-level Dictionary Learning--%%

partTrainGal = zeros(size(patchTrainGal,1)*ny,size(patchTrainGal,2)/nx);
for i = 1:size(partTrainGal,2)/ny
    tmp = patchTrainGal(:,(i-1)*nynx+1:i*nynx);
    tmp2 = zeros(672*ny, nx);
    for j = 1:ny
        tmp2((j-1)*672+1:j*672, :) = tmp(:, (j-1)*nx+1:j*nx);
    end
    partTrainGal(:,(i-1)*nx+1:i*nx) = tmp2;
end
partTrainPrb = zeros(size(patchTrainPrb,1)*ny,size(patchTrainPrb,2)/nx);
for i = 1:size(partTrainPrb,2)/ny
    tmp = patchTrainPrb(:,(i-1)*nynx+1:i*nynx);
    tmp2 = zeros(672*ny, nx);
    for j = 1:ny
        tmp2((j-1)*672+1:j*672, :) = tmp(:, (j-1)*nx+1:j*nx);
    end
    partTrainPrb(:,(i-1)*nx+1:i*nx) = tmp2;
end
partTestGal = zeros(size(testGal,1)*ny,size(testGal,2)/nx);
for i = 1:size(partTestGal,2)/ny
    tmp = testGal(:,(i-1)*nynx+1:i*nynx);
    tmp2 = zeros(672*ny, nx);
    for j = 1:ny
        tmp2((j-1)*672+1:j*672, :) = tmp(:, (j-1)*nx+1:j*nx);
    end
    partTestGal(:,(i-1)*nx+1:i*nx) = tmp2;
end
partTestPrb = zeros(size(testPrb,1)*ny,size(testPrb,2)/nx);
for i = 1:size(partTestPrb,2)/ny
    tmp = testPrb(:,(i-1)*nynx+1:i*nynx);
    tmp2 = zeros(672*ny, nx);
    for j = 1:ny
        tmp2((j-1)*672+1:j*672, :) = tmp(:, (j-1)*nx+1:j*nx);
    end
    partTestPrb(:,(i-1)*nx+1:i*nx) = tmp2;
end
clear patchTrainGal patchTrainPrb testGal testPrb

options=[];
options.ReducedDim = 300;
[Pca1,eigvalue] = PCA(partTrainGal',options);
Y1 = partTrainGal'*Pca1;

options=[];
options.ReducedDim = 300;
[Pca2,eigvalue2] = PCA(partTrainPrb',options);
Y2 = partTrainPrb'*Pca2;

clear partTrainGal partTrainPrb

Y1 = normcol_equal(Y1');
Y2 = normcol_equal(Y2');

opt = [];
opt.dsize = 50;
opt.lambda1 = 0.7;
opt.lambda2 = 0.001;
opt.lambda3 = 0.001;
opt.gamma = 1e-4;
opt.iter = 30;
opt.lambda4 = 0.1;

R = ones(size(Y1,2)) ./ size(Y1,2);
for i = 1:size(R,2)
    R(i,i) = 0;
end
opt.R = R;

%%--Part-level and image-level use the same function--%%
[DH1, DH2] = myDPL_ImageLevel_Dis(Y1, Y2, opt);

DH1 = normcol_equal(DH1);
DH2 = normcol_equal(DH2);

%%--Evaluation--%%%
Yte_galh = projTestData(DH1,Pca1'*partTestGal,opt.dsize);
Yte_prbh = projTestData(DH2,Pca2'*partTestPrb,opt.dsize);
clear partTestGal partTestPrb

partLevelDist = zeros(gsize, gsize);
for i = 1:gsize
    for j = 1:gsize
        partLevelDist(i,j) = (Yte_galh(:,i)'*Yte_prbh(:,j))/(norm(Yte_galh(:,i))*norm(Yte_prbh(:,j)));
    end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%

patchLevelDist2 = patchLevelDist / norm(patchLevelDist, 'fro');
imageLevelDist2 = imageLevelDist / norm(imageLevelDist, 'fro');
partLevelDist2 = partLevelDist / norm(partLevelDist, 'fro');

pwdist = -1*(patchLevelDist2 + 0.8*imageLevelDist2 + 3*partLevelDist2);
CMCC = evaluate_pwdist(pwdist)';

fprintf('CMC-rank1:%2.2f%%\n', CMCC(1)*100);
fprintf('CMC-rank5:%2.2f%%\n', CMCC(5)*100);
fprintf('CMC-rank10:%2.2f%%\n', CMCC(10)*100);
fprintf('CMC-rank20:%2.2f%%\n', CMCC(20)*100);